/* fft.c - Implement the Fast Fourier
 *         Transformation.  */

#include <math.h>
#include "fft.h"

/* Pi as a `long double' number */
#define PI_L 3.1415926535897932385L


unsigned int
next_power_of_two (unsigned int x)
/* Returns the smallest power of two,
 * which is greater than or equal to `x'.
 * This is a possible length of an input
 * vector for the function `fft'.  */
{
  int  result = 1;

  x -= 1;
  while (x>0)  x >>= 1, result <<=1;
  return  result;
}

static number
complex_exp (skalar phi)
/* Calculate `exp(i*phi)'.  */
{
  return  cosl(phi) + (0+1i)*sinl(phi);
}

static void
do_fft (number *data, unsigned int size,
        int reverse_flag)
/* Perform a fast (forward or reverse)
 * Fourier transformation on the vector
 * `data'.  The result is stored in place
 * of the input data.  `size' is the
 * length of the vector `data'.  If
 * `reverse_flag' is non-null, a reverse
 * Fourier transformation is done.  */
{
  unsigned int  m = size/2;
  number  x[m], y[m];
  unsigned int  j, k;
  number  q, q0;

  if (size>1) {
    for (j=1,k=0; j<size; j+=2)
      x[k++] = data[j];
    for (j=0,k=0; j<size; j+=2)
      y[k++] = data[j];
    do_fft (x, m, reverse_flag);
    do_fft (y, m, reverse_flag);
  
    if (reverse_flag) {
      q0 = complex_exp (2*PI_L/size);
    } else {
      q0 = complex_exp (-2*PI_L/size);
    }
    q = q0;
    for (j=0; j<m; ++j) {
      data [j] = x[j] + q*y[j];
      data [j+m] = x[j] - q*y[j];
      q *= q0;
    }
  }
}

void
fft (number *data, unsigned int size)
/* Perform a Fast Fourier Transformation
 * on the vector `data'.  The result is
 * stored in place of the input data.
 * `size' is the length of the vector
 * `data' and has to be a power of two. */
{
  unsigned int  j;
  number  q;

  do_fft (data, size, 0);
  q = sqrtl (1.0/size);
  for (j=0; j<size; ++j)  data[j] *= q;
}

void
reverse_fft (number *data,
             unsigned int size)
/* Perform a fast reverse Fourier
 * Transformation on the vector `data'.
 * The result is stored in place of the
 * input data.  `size' is the length of
 * the vector `data' and has to be a power
 * of two.  */
{
  unsigned int  j;
  number  q;

  do_fft (data, size, 1);
  q = sqrtl (1.0/size);
  for (j=0; j<size; ++j)  data[j] *= q;
}
