/* times.c - Geschwindigkeitsvergleich */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

#include "fft.h"

static void
method_a (unsigned int m, const number *f,
          const number *g, number *result)
/* Methode a: "Direkte" Berechnung der
 * Faltung.  `f' und `g' sind Vektoren der
 * Laenge `m'.  Die Funktion legt das
 * Ergebnis in `result' ab; dies sollte
 * ein Vektor der Groesse `2*m' sein.  */
{
  unsigned int  k, l;
  
  for (k=0; k<m; ++k) {
    number  zwerg = 0;
    for (l=0; l<k; ++l)
      zwerg += f[l] * g[k-l-1];
    result [k] = zwerg;
  }
  for (k=m; k<2*m; ++k) {
    number  zwerg = 0;
    for (l=k-m; l<m; ++l)
      zwerg += f[l] * g[k-l-1];
    result [k] = zwerg;
  }
}

static void
method_b (unsigned int m, const number *f,
          const number *g, number *result)
/* Methode b: Berechnet erst die
 * (`next_power_of_two(2*m)'-komponentige)
 * Fouriertransformierte von `data',
 * quadriert das Ergebnis und transformiert
 * es zurueck.  Das Ergebnis ist ein Vektor
 * der Laenge `next_power_of_two(2*m)' und
 * wird in `result' abgelegt.  */
{
  int  size = next_power_of_two(2*m);
  number  buf_f[size], buf_g[size];
  skalar  q = sqrtl (size);
  int  j;

  /* Kopiere die Eingabevektoren und
   * fuelle mit Nullen auf.  */
  memcpy (buf_f, f, m*sizeof(number));
  memcpy (buf_g, g, m*sizeof(number));
  for (j=m; j<size; ++j) {
    buf_f[j] = 0;  buf_g[j] = 0;
  }

  fft (buf_f, size);
  fft (buf_g, size);
  for (j=0; j<size; ++j)
    result[j] = q * buf_f[j] * buf_g[j];
  reverse_fft (result, size);
}

static void
doit (unsigned int m, const number *f,
      const number *g)
/* Misst die Ausfuehrungszeit beider
 * Methoden.  */
{
  int  size = next_power_of_two (2*m);
  number  result_a[2*m], result_b[size];
  clock_t  t1, t2, t3;
  double  time_a, time_b, base;
  skalar  delta;
  unsigned int  j, count;

  j = next_power_of_two (m);

  /* "Willkuerlich" festgelegte Anzahl
   * von Testlaeufen.  */
  count = (int)(80000.0 / (m*m));
  if (count>100)  count = 100;
  if (count<4)  count = 4;

  t1 = clock ();
  for (j=0; j<count; ++j)
    method_a (m, f, g, result_a);
  t2 = clock ();
  for (j=0; j<count; ++j)
    method_b (m, f, g, result_b);
  t3 = clock ();
  base = CLOCKS_PER_SEC*count;
  time_a = (t2-t1) / base;
  time_b = (t3-t2) / base;

  delta = 0;
  for (j=0; j<2*m; ++j) {
    number  x = result_b[j] - result_a[j];
    skalar  tmp = sqrtl (__real__(x * ~x));
    if (tmp>delta)  delta = tmp;
  }
  /* Die restlichen Komponenten sollten
   * wieder Null sein.  */
  for (j=m*2; j<size; ++j) {
    number  x = result_b[j];
    skalar  tmp = (__real__(x * ~x));
    if (tmp>delta)  delta = tmp;
  }

  printf ("%3u %6.1f %6.1f %12Lg %d\n",
          m, 1000*time_a, 1000*time_b,
          delta, count);
}

int
main ()
{
  unsigned int  j, m;
  number  f[300], g[300];

  /* Zufaellige Vektoren als Eingabe.  */
  for (j=0; j<300; ++j) {
    f[j] = (double)rand() / RAND_MAX;
    g[j] = (double)rand() / RAND_MAX;
  }
  
  printf ("  m  time_a time_b"
          "      Fehler Anzahl\n");
  for (m=2; m<=300; ++m)  doit (m, f, g);
  
  return  0;
}
